// © 2023, CGVR (https://cgvr.informatik.uni-bremen.de/),
// Author: Andre Mühlenbrock (muehlenb@uni-bremen.de)

#pragma once

#include "src/math/Mat4f.h"
#include "src/raytracer/Ray.h"

/**
 * Represents an intersection interval where the CSG primitive and
 * the ray overlaps each other.
 *
 * The naming of the attributes is identical to the attributes in
 * 'TracingContext', whereas 'a' names the first intersection and
 * 'b' names the second intersection.
 */

struct CSGInterval {
    // Intersection point data where the ray enters the object:
    float         aT;
    Vec4f         aHitPoint;
    Vec4f         aNormal;

    // Intersection point data where the ray leaves the object:
    float         bT;
    Vec4f         bHitPoint;
    Vec4f         bNormal;
};

/**
 * Represents a CSG Primitive.
 *
 * Note: Because we need all intersection points with CSGPrimitives
 * to determine the intersection point of the hole CSGTree, we need
 * another API for this primitives (instead the old Surface-API of
 * this framework).
 */

struct CSGPrimitive {
    virtual ~CSGPrimitive() = default;

    /**
     * Returns all intersections of the ray with the CSG Primitive as std::vector.
     */

    virtual std::vector<CSGInterval> intersectPrimitive(Ray &ray) const = 0;

    /**
     * Transformation matrix of this object (for translation, rotation and scale).
     *
     * This is set in the raytracingwidget.cpp when loading a CSGTree.
     */

    Mat4f transformation;

    /**
     * Normal matrix of this object.
     *
     * This is set in the raytracingwidget.cpp when loading a CSGTree.
     */

    Mat4f normalMatrix;

    /**
     * Inverse transformation matrix of this object.
     *
     * This is set in the raytracingwidget.cpp when loading a CSGTree.
     */

    Mat4f inverseTransform;
};

/**
 * A Sphere (CSGPrimitive).
 */

struct CSGSphere : CSGPrimitive
{
    /**
     * Performs an intersection test against the sphere.
     */

    virtual std::vector<CSGInterval> intersectPrimitive(Ray& ray) const override;
};

/**
 * A Cube (CSGPrimitive).
 */

struct CSGCube : CSGPrimitive
{
    /**
     * Calculates the normal in model space.
     */

    Vec4f calculateNormalMS(Vec4f dir) const;

    /**
     * Performs an intersection test against the cube.
     */

    virtual std::vector<CSGInterval> intersectPrimitive(Ray& ray) const override;
};


