
#include "CSGTree.h"

#include <iostream>

/**
 * Constructs the object and sets the root node.
 */

CSGTree::CSGTree(CSGNode* rootNode, Shader pShader)
    : rootNode(rootNode)
{
    shader = pShader;
}

/**
 * Destructor of the CSG Tree
 * (goes through all nodes and deletes the memory).
 */

CSGTree::~CSGTree()
{
    std::vector<CSGNode*> nodesToDelete;
    nodesToDelete.push_back(rootNode);

    while(!nodesToDelete.empty()){
        CSGNode* node = nodesToDelete.front();
        nodesToDelete.erase(nodesToDelete.cbegin());

        if(node->left != nullptr)
            nodesToDelete.push_back(node->left);

        if(node->right != nullptr)
            nodesToDelete.push_back(node->right);

        if(node->primitive != nullptr)
            delete node->primitive;

        delete node;
    }
}

/**
 * Tests for intersection between this CSGTree and the given
 * clRay in the section [t0, t1] and returns true if an inter-
 * section was found, otherwise false.
 *
 * Furthermore, the tContext variable is filled with information
 * about the intersection point. Variables which will be be
 * assigned are:
 *
 *      t          - Ray parameter at the intersection point
 *		v3Normal   - Normal vector of the intersection point
 *		v3HitPoint - Coordinate of the intersection point
 *		pclShader  - Pointer to the material (use getShader();)
 */

Intersection CSGTree::intersect(Ray ray, float tStart, float tEnd)
{
    std::vector<CSGInterval> intersectionIntervals = intersectNode(ray, rootNode);

    double minT = std::numeric_limits<double>::max();

    Intersection result;
    for(CSGInterval& interval : intersectionIntervals){
        if(interval.aT >= tStart && interval.aT < tEnd && interval.aT < minT){
            result.hasIntersection = true;
            result.t = interval.aT;
            result.normal = interval.aNormal;
            result.hitPoint = interval.aHitPoint;
            result.shader = &shader;
            minT = interval.bT;
        }

        if(interval.bT >= tStart && interval.bT < tEnd && interval.bT < minT){
            result.hasIntersection = true;
            result.t = interval.bT;
            result.normal = interval.bNormal;
            result.hitPoint = interval.bHitPoint;
            result.shader = &shader;
            minT = interval.bT;
        }
    }

    return result;
}

/**
 * Implements the union operation for merging intersection
 * intervals when recursively traverse the CSG tree.
 *
 * @param left  Intersection intervals of the left node.
 * @param right Intersection intervals of the right node.
 *
 * @return Intersection intervals after merging both nodes.
 */

std::vector<CSGInterval> CSGTree::csgUnion
    (
        std::vector<CSGInterval> left,
        std::vector<CSGInterval> right
        )
    const
{
    // ========================================================
    // TODO:
    //
    // Implement this method so that the union operation is
    // performed.
    // ========================================================

    std::vector<CSGInterval> result;

    return result;
}

/**
 * Implements the intersection operation for merging intersection
 * intervals when recursively traverse the CSG tree.
 *
 * @param left  Intersection intervals of the left node.
 * @param right Intersection intervals of the right node.
 *
 * @return Intersection intervals after merging both nodes.
 */

std::vector<CSGInterval> CSGTree::csgIntersection
    (
        std::vector<CSGInterval> left,
        std::vector<CSGInterval> right
        )
    const
{
    // ========================================================
    // TODO:
    //
    // Implement this method so that the intersection operation
    // is performed.
    // ========================================================

    std::vector<CSGInterval> result;

    return result;
}

/**
 * Implements the difference operation for merging intersection
 * intervals when recursively traverse the CSG tree.
 *
 * @param left  Intersection intervals of the left node.
 * @param right Intersection intervals of the right node.
 *
 * @return Intersection intervals after merging both nodes.
 */

std::vector<CSGInterval> CSGTree::csgDifference
    (
        std::vector<CSGInterval> left,
        std::vector<CSGInterval> right
        )
    const
{
    // ========================================================
    // TODO:
    //
    // Implement this method so that the difference operation
    // is performed.
    // ========================================================

    std::vector<CSGInterval> result;

    return result;
}

/**
 * Returns all intersections of the current node.
 */

std::vector<CSGInterval> CSGTree::intersectNode(Ray &clRay, CSGNode* node) const
{
    if(node->type == LEAF_NODE){
        return node->primitive->intersectPrimitive(clRay);
    }

    if(node->type == OP_UNION){
        return csgUnion(
            intersectNode(clRay, node->left),
            intersectNode(clRay, node->right)
        );
    }

    if(node->type == OP_INTERSECTION){
        return csgIntersection(
            intersectNode(clRay, node->left),
            intersectNode(clRay, node->right)
        );
    }

    if(node->type == OP_DIFFERENCE){
        return csgDifference(
            intersectNode(clRay, node->left),
            intersectNode(clRay, node->right)
        );
    }

    return std::vector<CSGInterval>();
}

/**
 * Prints the CSGTree to console for debugging purposes.
 */

void CSGTree::print(){
    std::cout << "CSG Tree" << std::endl;
    print(rootNode, 1);
}

/**
 * Prints a node of the CSGTree recursively.
 */

void CSGTree::print(CSGNode* node, int level){
    std::string preTabs = "";
    for(int i=0; i < level; ++i)
        preTabs += "\t";

    std::cout << preTabs.c_str() << "Node Type: " << node->type << std::endl;
    std::cout << preTabs.c_str() << "Primitive: " << node->primitive << std::endl;

    if(node->left != nullptr)
        print(node->left, level + 1);

    if(node->right != nullptr)
        print(node->right, level + 1);
}
