// © 2023, CGVR (https://cgvr.informatik.uni-bremen.de/),
// Author: Andre Mühlenbrock (muehlenb@uni-bremen.de)

#pragma once

#include "src/scene/Surface.h"
#include "src/scene/surfaces/CSGPrimitives.h"


/**
 * Represents a CSGTree.
 */

class CSGTree : public Surface
{

public:
    /**
     * Represents the type as well as the operation of a CSGNode, if any.
     *
     * - If the type is LEAF_NODE, the CSGNode is a leaf node and only
     * contains a leaf (CSGPrimitive).
     * - If the type is OP_UNION, OP_INTERSECTION or OP_DIFFERENCE, the
     * node is an intermediate node and contains a left and right child
     * node
     */

    enum NodeType { LEAF_NODE, OP_UNION, OP_INTERSECTION, OP_DIFFERENCE};


    /**
     * Represents a node in a csg tree.
     *
     * For simplicity, this node can be either:
     * - an intermediate node in the tree, which has two child nodes
     *   (left and right), and an node type != LEAF_NODE
     * - a leaf node with a primitive != NONE.
     */

    struct CSGNode
    {
        /**
         * Constructor that simply set all attributes.
         */

        CSGNode(NodeType type, CSGNode* left, CSGNode* right, CSGPrimitive* primitive)
            : type(type), left(left), right(right), primitive(primitive){};

        // Type of the node (e.g. LEAF_NODE or type of operation):
        NodeType type;

        // Left child (if it's NOT a LEAF_NODE, otherwise nullptr):
        CSGNode* left;

        // Right child (if it's NOT a LEAF_NODE, otherwise nullptr):
        CSGNode* right;

        // Type of primitive (if it's a LEAF NODE, otherwise nullptr):
        CSGPrimitive* primitive;
    };

    /**
     * Constructs the object and sets the root node.
     */

    CSGTree(CSGNode* rootNode, Shader shader);

    /**
     * Destructor for the CSGTree.
     */

    ~CSGTree();

    /**
     * Performs an intersection test between the given ray and the
     * CSGTree in the interval [t0, t1]. If an intersection is found,
     * this function returns true and inserts the hitpoint, normal and
     * t of intersection into the given tracing context.
     */
    virtual Intersection intersect(Ray ray, float tStart, float tEnd) override;

    /**
     * Performs an intersection test between the node (including all
     * child nodes) and the given ray recursively.
     */

    std::vector<CSGInterval> intersectNode(Ray &clRay, CSGNode* node) const;

    /**
     * Implements the union operation for merging intersection intervals
     * when recursively traverse the CSG tree.
     *
     * @param left  Intersection intervals of the left node.
     * @param right Intersection intervals of the right node.
     *
     * @return Intersection intervals after merging both nodes.
     */

    std::vector<CSGInterval> csgUnion(std::vector<CSGInterval> left, std::vector<CSGInterval> right) const;

    /**
     * Implements the intersection operation for merging intersection
     * intervals when recursively traverse the CSG tree.
     *
     * @param left  Intersection intervals of the left node.
     * @param right Intersection intervals of the right node.
     *
     * @return Intersection intervals after merging both nodes.
     */

    std::vector<CSGInterval> csgIntersection(std::vector<CSGInterval> left, std::vector<CSGInterval> right) const;

    /**
     * Implements the difference operation for merging intersection
     * intervals when recursively traverse the CSG tree.
     *
     * @param left  Intersection intervals of the left node.
     * @param right Intersection intervals of the right node.
     *
     * @return Intersection intervals after merging both nodes.
     */

    std::vector<CSGInterval> csgDifference(std::vector<CSGInterval> left, std::vector<CSGInterval> right) const;

    /**
     * Prints the CSGTree to console for debugging purposes.
     */

    void print();

private:
    /**
     * The root node of the CSG Tree.
     */

    CSGNode* rootNode;

    /**
     * Print the given node with all child nodes recursively for debugging purposes.
     */

    void print(CSGNode* node, int level);
};
